# Copyright (c) 2024-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import torch


class CompressedLinear(torch.autograd.Function):
    @staticmethod
    def forward(ctx, inputs, _p, bias):
        # manual cast
        ctx._inputs_dtype = inputs.dtype
        if torch.is_autocast_enabled():
            inputs = inputs.to(torch.get_autocast_gpu_dtype())
        output = _p.linear(inputs, bias)
        ctx.save_for_backward(inputs, bias)
        ctx.carry = (_p,)
        return output

    @staticmethod
    def backward(ctx, grad_output):
        (inputs, bias) = ctx.saved_tensors
        (_p,) = ctx.carry

        inputs_shape = inputs.shape

        inputs = inputs.reshape(-1, inputs.shape[-1])
        grad_output = grad_output.reshape(-1, grad_output.shape[-1])
        _weight = _p()

        grad_inputs = torch.mm(grad_output, _weight).reshape(inputs_shape).to(ctx._inputs_dtype)
        _weight = torch.mm(grad_output.t(), inputs)

        if _p._gradient is not None:
            _p._gradient.add_(_weight)
        else:
            _p._gradient = _weight

        return (
            grad_inputs,
            torch.empty_like(_p.data),
            grad_output.sum(0) if bias is not None else None,
        )



# Copyright (c) 2024-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


from neuzip.ops.linear import CompressedLinear

__all__ = ["CompressedLinear"]



# Copyright (c) 2024-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import torch


class Parameter(torch.nn.Parameter):
    def __new__(cls, manager, data, requires_grad=True) -> "Parameter":
        p = super().__new__(cls, torch.zeros(0).to(data), requires_grad)

        p._handle = manager.write(data)
        p._gradient = None
        p._manager = manager
        p._shape = data.shape

        del data
        return p

    @torch.no_grad()
    def __call__(self) -> torch.Tensor:
        return self._manager.read(self._handle).reshape(self._shape).to(self.device).requires_grad_(False)

    def linear(self, inputs, bias=None):
        return self._manager.linear(self._handle, inputs, bias).to(self.device)



# Copyright (c) 2024-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


from neuzip.nn.embedding import Embedding
from neuzip.nn.linear import Linear
from neuzip.nn.module import Parameter

__all__ = ["Parameter", "Linear", "Embedding"]



# Copyright (c) 2024-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import torch

from neuzip.nn.module import Parameter


class Embedding(torch.nn.Module):
    def __init__(self, manager, embedding: torch.nn.Embedding) -> None:
        super().__init__()
        self.manager = manager
        self.num_embeddings = embedding.num_embeddings
        self.embedding_dim = embedding.embedding_dim
        self.padding_idx = embedding.padding_idx
        self.max_norm = embedding.max_norm
        self.norm_type = embedding.norm_type
        self.scale_grad_by_freq = embedding.scale_grad_by_freq
        self.sparse = embedding.sparse

        self._p = Parameter(self.manager, embedding.weight)

    def extra_repr(self) -> str:
        return f"! num_embeddings={self.num_embeddings}, embedding_dim={self.embedding_dim}, padding_idx={self.padding_idx}, max_norm={self.max_norm}, norm_type={self.norm_type}, scale_grad_by_freq={self.scale_grad_by_freq}, sparse={self.sparse}"  # noqa

    def forward(self, x):
        return torch.nn.functional.embedding(
            x,
            self._p(),
            self.padding_idx,
            self.max_norm,
            self.norm_type,
            self.scale_grad_by_freq,
            self.sparse,
        )



# Copyright (c) 2024-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import torch
import transformers

from neuzip.nn.module import Parameter
from neuzip.ops import CompressedLinear


class Conv1D(torch.nn.Module):
    def __init__(self, manager, linear: transformers.pytorch_utils.Conv1D) -> None:
        super().__init__()
        self.manager = manager
        self.in_features, self.out_features = linear.weight.shape
        self._p = Parameter(self.manager, linear.weight.data)
        if linear.bias is not None:
            self.bias = torch.nn.Parameter(linear.bias.data)
        else:
            self.register_parameter("bias", None)

    def extra_repr(self) -> str:
        return f"! in_features={self.in_features}, out_features={self.out_features}, bias={self.bias is not None}"

    def forward(self, x):
        return CompressedLinear.apply(x, self._p, self.bias)



# Copyright (c) 2024-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import logging

import torch
import transformers

from neuzip.nn import Embedding, Linear

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)


def _set_module_by_name(parent_module, name, child_module):
    module_names = name.split(".")
    if len(module_names) == 1:
        setattr(parent_module, name, child_module)
    else:
        next_module = parent_module
        for idx in range(len(module_names) - 1):
            next_module_name = module_names[idx]
            if next_module_name.isnumeric():
                next_module = next_module[int(next_module_name)]
            else:
                next_module = getattr(next_module, next_module_name)
        setattr(next_module, module_names[-1], child_module)


def _delete_module(module):
    for _name, _param in module.named_parameters():
        # Save device memory by clearing parameters
        setattr(module, _name, None)
        del _param


def _unroll_convert(manager, module):
    # Quantization happens in-place
    for name, m in module.named_modules():
        if isinstance(m, torch.nn.Linear):
            full_name = f"{name}"
            logger.debug(f"Quantizing {full_name}: {m}")
            _set_module_by_name(module, name, Linear(manager, m))
            _delete_module(m)
        elif isinstance(m, torch.nn.Embedding):
            full_name = f"{name}"
            logger.debug(f"Quantizing {full_name}: {m}")
            _set_module_by_name(module, name, Embedding(manager, m))
            _delete_module(m)
    return module


def _convert(manager, module, prefix=""):
    for name, child in module.named_children():
        if "lm_head" in name:
            pass
        elif isinstance(child, (torch.nn.Linear, transformers.pytorch_utils.Conv1D)):
            full_name = f"{prefix}.{name}" if prefix else name
            logger.debug(f"Converting {full_name}: {child}")
            setattr(module, name, Linear(manager, child))
            _delete_module(child)
        else:
            _convert(manager, child, f"{prefix}.{name}" if prefix else name)
    return module



# Copyright (c) 2024-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import os

import torch
from setuptools import find_packages, setup
from torch.utils.cpp_extension import BuildExtension, CUDAExtension

NEUZIP_VERSION = "0.1.0"

setup(
    name="neuzip",
    ext_modules=[
        CUDAExtension(
            name="neuzip._cuda",
            sources=[
                "cuda/neuzip.cu",
            ],
            include_dirs=[
                os.path.dirname(os.path.abspath(__file__)),
                os.path.join(os.path.dirname(os.path.abspath(__file__)), "cuda"),
            ],
            libraries=["nvcomp"],
            extra_compile_args=[
                "-w",
                "-O3",
                f"-D_GLIBCXX_USE_CXX11_ABI={int(torch._C._GLIBCXX_USE_CXX11_ABI)}",
            ],
        ),
    ],
    install_requires=["torch", "transformers", "numpy"],
    packages=find_packages(),
    package_data={
        "neuzip._cuda": ["*.cuh", "*.cu"],
    },
    include_package_data=True,
    url="https://github.com/BorealisAI/neuzip",
    license="CC BY-NC-SA 4.0",
    version=NEUZIP_VERSION,
    cmdclass={"build_ext": BuildExtension},
)



# Copyright (c) 2024-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import torch

from neuzip import nn, ops
from neuzip.binding import Algorithm, Manager

__all__ = ["ops", "nn", "Algorithm", "Manager"]


def lomo_hook(scheduler, param):
    if param.grad is None:
        return
    learning_rate = scheduler.get_last_lr()[0]
    with torch.no_grad():
        param.data.add_(param.grad, alpha=-learning_rate)
    del param.grad


def compressed_lomo_hook(scheduler, param):
    if param._gradient is None:
        return
    learning_rate = scheduler.get_last_lr()[0]
    with torch.no_grad():
        weight = param()
        weight.add_(param._gradient, alpha=-learning_rate)
        param._manager.write(weight, param._handle)
    del param._gradient
    del param.grad
    param._gradient = None



# Copyright (c) 2024-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import enum
from typing import Optional

import torch

from neuzip.utils import _convert

try:
    import neuzip._cuda
except ImportError:
    raise ImportError(
        "The CUDA extension is not built. Please run `python setup.py install`"
    )


class Algorithm(enum.Enum):
    ans = neuzip._cuda.Algorithm.ans
    gdeflate = neuzip._cuda.Algorithm.gdeflate
    zstd = neuzip._cuda.Algorithm.zstd
    dummy = "dummy"


@torch.no_grad()
def _dummy_linear(weights, inputs: torch.Tensor, bias=None) -> torch.Tensor:
    return torch.nn.functional.linear(inputs, weights, bias)


class DummyBackend:
    def write(self, handle: str, tensor: torch.Tensor) -> None:
        cuda_tensor = tensor.cuda()
        del tensor
        setattr(self, handle, cuda_tensor)

    def read(self, handle: str) -> torch.Tensor:
        return getattr(self, handle)

    def size(self, handle: str) -> int:
        result = getattr(self, handle)
        return result.numel() * result.element_size()

    def linear(
        self, handle: str, inputs: torch.Tensor, shape: torch.Size, bias=None
    ) -> torch.Tensor:
        weights = self.read(handle).reshape(shape)
        return _dummy_linear(weights, inputs, bias)


class Manager:
    def __init__(
        self,
        algorithm: Algorithm = Algorithm.ans,
        precision: int = 7,
        normalizer_size=None,
        chunk_size=2**16,
    ) -> None:
        if normalizer_size is None:
            normalizer_size = 0

        if algorithm == Algorithm.dummy:
            self._be = DummyBackend()
        else:
            module_name = f"Manager_f{precision}_n{normalizer_size}"
            self._be = getattr(neuzip._cuda, module_name)(algorithm.value, chunk_size)

        self.meta_dict = {}

    def write(self, tensor: torch.Tensor, handle: Optional[str] = None) -> str:
        if handle is None:
            handle = str(len(self.meta_dict))

        oshape = tensor.shape

        self.meta_dict[handle] = oshape

        self._be.write(handle, tensor)

        del tensor

        return handle

    def read(self, handle: str) -> torch.Tensor:
        shape = self.meta_dict[handle]
        tensor = self._be.read(handle).requires_grad_(False)
        tensor = tensor.reshape(shape)
        return tensor

    def size(self, handle: str) -> int:
        return self._be.size(handle)

    def linear(self, handle: str, inputs: torch.Tensor, bias=None) -> torch.Tensor:
        shape = self.meta_dict[handle]
        if bias is not None:
            return self._be.linear(handle, inputs, shape, bias)
        else:
            return self._be.linear_without_bias(handle, inputs, shape)

    # additional methods
    def convert(self, module: torch.nn.Module) -> torch.nn.Module:
        return _convert(self, module)

    def split(self, tensor: torch.Tensor) -> str:
        shape = tensor.shape
        exp, frac = self._be.split(tensor)
        return exp.reshape(shape), frac.reshape(shape)

    def __del__(self):
        del self._be
        del self.meta_dict
